/**
 * Copyright (c) FOM-Nikhef 2014-
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *
 *  Authors:
 *  2014-
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl>
 *
 */

#define _XOPEN_SOURCE	500

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pwd.h>

#include <openssl/x509.h>

#include "lcmaps_plugins_robot_config.h"

#include <lcmaps/lcmaps_modules.h>
#include <lcmaps/lcmaps_arguments.h>
#include <lcmaps/lcmaps_cred_data.h>

#if defined(HAVE_LCMAPS_LCMAPS_PLUGIN_PROTOTYPES_H)
#   include <lcmaps/lcmaps_plugin_prototypes.h>
#else
#   include "lcmaps_plugin_prototypes.h"
#endif

#include "lcmaps_gridmapfile.h"
#include "lcmaps_subuserdn.h"


/************************************************************************
 * defines
 ************************************************************************/

#define PLUGIN_PREFIX	"lcmaps_robot_ban_dn"

#define PLUGIN_RUN	0   /* full run mode */
#define PLUGIN_VERIFY	1   /* verify-only mode */


/************************************************************************
 * global variables
 ************************************************************************/

static char *banmapfile = NULL; /* Filename of banning file */
static int  use_wildcard = 1;    /* default use wildcards */


/************************************************************************
 * private prototypes
 ************************************************************************/

/* called by plugin_run() and plugin_verify() */
static int plugin_run_or_verify(int argc, lcmaps_argument_t *argv,
				int lcmaps_mode);


/************************************************************************
 * public functions
 ************************************************************************/

/**
 * Initialize function for plugin
 */
int plugin_initialize(int argc, char **argv) {
    const char * logstr = PLUGIN_PREFIX"-plugin_initialize()";
    int i;

    /* Log commandline parameters on debug */
    lcmaps_log(LOG_DEBUG,"%s: passed arguments:\n",logstr);
    for (i=0; i < argc; i++)
	lcmaps_log(LOG_DEBUG,"%s: arg %d is %s\n", logstr, i, argv[i]);

    /* Parse arguments, argv[0] = name of plugin, so start with i = 1 */
    for (i = 1; i < argc; i++) {
	/* check banmapfile option (or similar variations) */
        if ( strcmp(argv[i], "-banmapfile") == 0 ||
             strcmp(argv[i], "-BANMAPFILE") == 0 ||
             strcmp(argv[i], "-ban_mapfile") == 0 ||
             strcmp(argv[i], "-BAN_MAPFILE") == 0 ||
             strcmp(argv[i], "-gridmapfile") == 0 ||
             strcmp(argv[i], "-GRIDMAPFILE") == 0 ||
             strcmp(argv[i], "-gridmap") == 0 ||
             strcmp(argv[i], "-GRIDMAP") == 0 )
	{
	    /* check valid filename argument */
            if (argv[i + 1] == NULL || argv[i + 1][0]=='\0') {
		lcmaps_log(LOG_ERR,
		    "%s: option %s needs to be followed by valid filename\n",
		    logstr, argv[i]);
		goto fail_init_module;
	    }
	    /* free existing one and copy new one */
	    free(banmapfile); banmapfile=NULL;
	    if (argv[i+1][0]=='/')  { /* absolute path */
		if ( (banmapfile = strdup(argv[i + 1])) == NULL)	{
		    lcmaps_log(LOG_ERR, "%s: out of memory\n", logstr);
		    goto fail_init_module;
		}
	    } else { /* relative path */
		if (lcmaps_get_prefixed_file(argv[i + 1], &banmapfile)==-1)
		    goto fail_init_module;
	    }
	    /* log the name and increase arg counter */
	    lcmaps_log(LOG_DEBUG, "%s: Using ban file \"%s\".\n",
		    logstr, banmapfile);
            i++;
        }
	/* check whether to disable wildcard matching */
	else if (strcmp(argv[i], "-disablewildcard") == 0 ||
                 strcmp(argv[i], "--disablewildcard") == 0 ||
                 strcmp(argv[i], "-disablewildcards") == 0 ||
                 strcmp(argv[i], "--disablewildcards") == 0 ||
                 strcmp(argv[i], "-no_wildcard") == 0)
	{
	    lcmaps_log(LOG_DEBUG, "%s: disabling wildcard matching\n", logstr);
            use_wildcard = 0;
        }
	/* any other argument is an error */
	else
	{
            lcmaps_log(LOG_ERR,
		    "%s: Unknown argument for plugin: %s (failure)\n",
		    logstr, argv[i]);
            goto fail_init_module;
        }
    }

    /* Check we have a banmapfile */
    if (banmapfile==NULL || banmapfile[0]=='\0')    {
        lcmaps_log(LOG_ERR,
	    "%s: Error: No banmapfile was provided. "
	    "Use the option -banmapfile <banmapfile> to set one.\n", logstr);
        goto fail_init_module;
    }

    return LCMAPS_MOD_SUCCESS;

fail_init_module:
    free(banmapfile);
    banmapfile = NULL;

    return LCMAPS_MOD_FAIL;
}

/**
 * Introspect function for plugin, defines which lcmaps framework data is made
 * available as commandline options for the plugin_run.
 */
int plugin_introspect(int *argc, lcmaps_argument_t **argv) {
    const char * logstr = PLUGIN_PREFIX"-plugin_introspect()";
    static lcmaps_argument_t argList[] = {
	{"px509_chain","STACK_OF(X509) *", 0,NULL},
	{"pem_string" ,"char *"          , 0,NULL},
	{NULL         ,NULL              ,-1,NULL}
    };

    lcmaps_log(LOG_DEBUG,"%s: introspecting\n", logstr);

    *argv = argList;
    *argc = lcmaps_cntArgs(argList);
    lcmaps_log(LOG_DEBUG,"%s: address first argument: %p\n",
	    logstr, (void*)argList);

    return LCMAPS_MOD_SUCCESS;
}

/**
 * run function for plugin, wrapper around plugin_run_or_verify() with
 * PLUGIN_RUN mode
 */
int plugin_run(int argc, lcmaps_argument_t *argv) {
    return plugin_run_or_verify(argc, argv, PLUGIN_RUN);
}

/**
 * run function for plugin, wrapper around plugin_run_or_verify() with
 * PLUGIN_VERIFY mode
 */
int plugin_verify(int argc, lcmaps_argument_t *argv) {
    return plugin_run_or_verify(argc, argv, PLUGIN_VERIFY);
}

/**
 * terminate plugin, frees used memory.
 * return LCMAPS_MOD_SUCCESS
 */
int plugin_terminate(void) {
    const char * logstr = PLUGIN_PREFIX"-plugin_terminate()";

    lcmaps_log(LOG_DEBUG,"%s: terminating\n", logstr);

    free(banmapfile);
    banmapfile=NULL;

    return LCMAPS_MOD_SUCCESS;
}


/************************************************************************
 * private functions
 ************************************************************************/

/**
 * Actual run/verify function. Called by both plugin_run and plugin_verify
 * with different lcmaps_mode.
 */
static int plugin_run_or_verify(int argc, lcmaps_argument_t *argv,
				int lcmaps_mode) {
    const char *        logstr       = NULL;
    void *              value        = NULL;
    STACK_OF(X509)*     chain        = NULL;
    char *              pem          = NULL;
    char *              dn           = NULL;
    unsigned short      options      = 0;
    int                 rc           = 0;
    char *              username     = NULL;

    /* Set suitable logstr */
    if (lcmaps_mode == PLUGIN_RUN)
        logstr = PLUGIN_PREFIX"-plugin_run()";
    else if (lcmaps_mode == PLUGIN_VERIFY)
        logstr = PLUGIN_PREFIX"-plugin_verify()";
    else {
        lcmaps_log(LOG_ERR, PLUGIN_PREFIX"-plugin_run_or_verify(): "
		"attempt to run plugin in invalid mode: %d\n", lcmaps_mode);
        goto fail_plugin;
    }

    /* Try to get chain from LCMAPS */
    value=lcmaps_getArgValue("px509_chain", "STACK_OF(X509) *", argc, argv);
    if (value == NULL || (chain = *(STACK_OF(X509) **)value) == NULL ) {
	/* No valid chain found in LCMAPS framework, try to obtain from PEM
	 * string */
	lcmaps_log(LOG_DEBUG, "%s: no X.509 chain is set, trying pem string.\n",
		logstr);
	value=lcmaps_getArgValue("pem_string", "char *", argc, argv);
	if (value==NULL || (pem=*(char**)value) == NULL ) {
	    /* also not found: fatal error */
	    lcmaps_log(LOG_WARNING,
		    "%s: no chain or pemstring is set.\n", logstr);
	    goto fail_plugin;
	}

	/* Convert pem string to chain */
	if (pem_string_to_x509_chain(&chain, pem)!=0)	{
	    lcmaps_log(LOG_WARNING,
		    "%s: cannot convert pemstring to chain.\n", logstr);
	    goto fail_plugin;
	}
    }

    /* Get per-user sub-proxy DN from certificate chain */
    if (lcmaps_get_subuser_dn(&dn, chain)==-1)   {
	lcmaps_log(LOG_INFO,
		"%s: cannot get per-user sub-proxy DN from chain\n", logstr);
	goto success_plugin;
    }

    /* Check if we have a banmapfile */
    if (banmapfile==NULL || banmapfile[0]=='\0')   {
        lcmaps_log(LOG_ERR, "%s: Error: No (valid) banmapfile assigned.\n",
		logstr);
        goto fail_plugin;
    }

    /* Log filename */
    lcmaps_log(LOG_DEBUG, "%s: banmapfile is: %s\n", logstr, banmapfile);

    /* Set the matching options */

    /* match DN only and ban-mapfile must exist */
    options = MATCH_ONLY_DN|REQUIRE_MAPFILE;

    /* check if we should use wildcards ... */
    if (use_wildcard)
	options = options|MATCH_WILD_CHARS;

    /* Do lookup for per-user sub-proxy DN in ban mapfile */
    rc = lcmaps_gridmapfile(banmapfile, dn, NULL, options, &username);

    /* parse return value of lcmaps_gridmapfile */
    if (rc==-1)	{
	/* error */
	lcmaps_log(LOG_ERR, "%s: error while looking up DN in banmapfile %s\n",
	    logstr, banmapfile);
	goto fail_plugin;
    }
    if (rc==1)	{
	/* match */
	lcmaps_log(LOG_NOTICE,
		"%s: DN \"%s\" is found on the ban list.\n", logstr, dn);
	goto fail_plugin;
    }

    /* did not find match */
    lcmaps_log(LOG_DEBUG,
	    "%s: No entry found in the ban list: user is NOT banned.\n",
	    logstr);

    /* success */
success_plugin:
    /* Cleanup chain but only if we allocated it */
    if (pem && chain)
	sk_X509_pop_free(chain, X509_free);
    free(dn);
    free(username);

    lcmaps_log(LOG_INFO,"%s: robot_ban_dn plugin succeeded\n", logstr);

    return LCMAPS_MOD_SUCCESS;

fail_plugin:
    /* Cleanup chain but only if we allocated it */
    if (pem && chain)
	sk_X509_pop_free(chain, X509_free);
    free(dn);
    free(username);

    lcmaps_log(LOG_INFO,"%s: robot_ban_dn plugin failed\n", logstr);

    return LCMAPS_MOD_FAIL;
}
